--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Parser = require("dogma.syn.Parser")
local SentType = require("dogma.syn.SentType")
local StmtType = require("dogma.syn.StmtType")

--Suite.
return suite("dogma.syn.StmtParser", function()
  local parser, stmt

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
  end):title("Create parser")

  -----------
  -- break --
  -----------
  suite("break", function()
    test("break", function()
      parser:parse("break")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop=nil})
    end)

    test("break for", function()
      parser:parse("break for")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="for"})
    end)

    test("break for each", function()
      parser:parse("break for each")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="for each"})
    end)

    test("break while", function()
      parser:parse("break while")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="while"})
    end)

    test("break until", function()
      parser:parse("break until")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.BREAK, loop="until"})
    end)

    test("break x - error", function()
      parser:parse("break x")
      assert(function() parser:next() end):raises("anonymous code (1, 7): end of line expected.")
    end)
  end)

  ----------------
  -- nextNext() --
  ----------------
  suite("nextNext()", function()
    test("next", function()
      parser:parse("next")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.NEXT})
    end)

    test("next x - error", function()
      parser:parse("next x")
      assert(function() parser:next() end):raises("anonymous code (1, 6): end of line expected.")
    end)
  end)

  ------------------
  -- nextReturn() --
  ------------------
  suite("nextReturn()", function()
    test("return", function()
      parser:parse("return")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.RETURN})
      assert(stmt):len(0)
    end)

    test("return x+y", function()
      parser:parse("return x+y")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.RETURN})
      assert(stmt):len(1)
      assert(stmt.values[1]:__tostring()):eq("(+ x y)")
    end)
  end)

  -----------------
  -- nextYield() --
  -----------------
  suite("nextYield()", function()
    test("yield x+y", function()
      parser:parse("yield x+y")
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.YIELD})
      assert(tostring(stmt.value)):eq("(+ x y)")
    end)
  end):tags("yield")

  ---------
  -- use --
  ---------
  suite("use", function()
    suite("template", function()
      test("use Template<T>", function()
        parser:parse('use Map<T>')
        stmt = parser:next()
        assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
        assert(stmt.modules):eq({})
      end)

      test("use Template<InvalidName>", function()
        parser:parse("use Map<G>")
        assert(function() parser:next() end):raises("anonymous code (1, 9): name expected: 'T'. Received: 'G'.")
      end)
    end)

    test("use <id> - lint", function()
      parser:parse('use <fs>')
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="fs", path="fs"}
      })
    end)

    test("use <id> - trans", function()
      parser:parse('use <fs>', nil, {imports={fs="@dogmalang/fs2"}})
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="fs", path="@dogmalang/fs2"}
      })
    end)

    test("use <id> - unknown alias", function()
      parser:parse('use <fs2>', nil, {imports={fs="@dogmalang/fs"}})
      assert(function() parser:next() end):raises("anonymous code (1, 5): unknown module alias: fs2.")
    end)

    test("use LiteralStr", function()
      parser:parse('use "module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module", path = "module"}
      })
    end)

    test("use\\n LiteralStr", function()
      parser:parse('use\n "module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module", path = "module"}
      })
    end)

    test("use\\nLiteralStr", function()
      parser:parse('use\n"module"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({})
    end)

    test('use "justo.assert"', function()
      parser:parse('use "justo.assert"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "assert", path = "justo.assert"}
      })
    end)

    test('use "redispark-connector-redis"', function()
      parser:parse('use "redispark-connector-redis"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "redis", path = "redispark-connector-redis"}
      })
    end)

    test('use "@scope/pkg"', function()
      parser:parse('use "@scope/pkg"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "pkg", path = "@scope/pkg"}
      })
    end)

    test('use "@scope/my-pkg"', function()
      parser:parse('use "@scope/my-pkg"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "pkg", path = "@scope/my-pkg"}
      })
    end)

    test("use LitealStr, LiteralStr", function()
      parser:parse('use "module1", "module2"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module1", path = "module1"},
        {name = "module2", path = "module2"}
      })
    end)

    test("use\\n  LiteralStr\\n  LiteralStr", function()
      parser:parse('use\n "module1"\n "module2"')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "module1", path = "module1"},
        {name = "module2", path = "module2"}
      })
    end)

    test("use LiteralStr as Name", function()
      parser:parse('use "module" as mod')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod", path = "module"}
      })
    end)

    test("use\\n LiteralStr as Name", function()
      parser:parse('use\n "module" as mod')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod", path = "module"}
      })
    end)

    test("use LiteralStr as Name, LiteralStr as Name", function()
      parser:parse('use "module1" as mod1, "module2" as mod2')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod1", path = "module1"},
        {name = "mod2", path = "module2"}
      })
    end)

    test("use\\n LiteralStr as Name\\n LiteralStr as Name", function()
      parser:parse('use\n "module1" as mod1\n "module2" as mod2')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "mod1", path = "module1"},
        {name = "mod2", path = "module2"}
      })
    end)

    test([[use "module1"! "module2" - error - comma expected]], function()
      parser:parse([[use "module1"! "module2"]])
      assert(function() parser:next() end):raises("anonymous code (1, 14): ',' or end of line expected.")
    end)

    test([[use "~module" - error - invalid module path format]], function()
      parser:parse([[use "~module"]])
      assert(function() parser:next() end):raises("invalid module path format: '~module'.")
    end)

    test([[use "one/two"]], function()
      parser:parse([[use "one/two"]])
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "two", path = "one/two"}
      })
    end)

    test("use Literal, Name", function()
      parser:parse('use "one", two')
      stmt = parser:next()
      assert(stmt):isTable():has({type = SentType.STMT, subtype = StmtType.USE})
      assert(stmt.modules):eq({
        {name = "one", path = "one"},
        {name = "two", path = "./two"}
      })
    end)

    test("use name", function()
      parser:parse("use file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="./file"}
      })
    end)

    test("use name => {variables}", function()
      parser:parse("use module => {var1,var2,var3}")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="module", path="./module", uVars={"var1", "var2", "var3"}}
      })
    end)

    test("use name/name", function()
      parser:parse("use dir/file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="./dir/file"}
      })
    end)

    test("use name/name/name", function()
      parser:parse("use dir/subdir/file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="./dir/subdir/file"}
      })
    end)

    test("use ../name", function()
      parser:parse("use ../file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="../file"}
      })
    end)

    test("use ../name/name", function()
      parser:parse("use ../dir/file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="../dir/file"}
      })
    end)

    test("use ../../name", function()
      parser:parse("use ../../file")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="file", path="../../file"}
      })
    end)

    test("use (\\nname1, name2\\nname3)", function()
      parser:parse("use (\nname1, name2\nname3\n)")
      stmt = parser:next()
      assert(stmt):isTable():has({type=SentType.STMT, subtype=StmtType.USE})
      assert(stmt.modules):eq({
        {name="name1", path="./name1"},
        {name="name2", path="./name2"},
        {name="name3", path="./name3"}
      })
    end)

    test("use Number - error", function()
      parser:parse("use 123")
      assert(function() parser:next() end):raises("anonymous code (1, 5): <name>, name, literal text or '..' expected.")
    end)
  end)

  ----------
  -- from --
  ----------
  suite("from", function()
    test("from LiteralStr use Name", function()
      parser:parse([[from "fs" use FSWatcher]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "FSWatcher"}
        }
      })
    end)

    test("from LiteralStr use Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"}
        }
      })
    end)

    test("from LiteralStr use Name, Name", function()
      parser:parse([[from "fs" use FSWatcher, watch]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "FSWatcher"},
          {name = "watch", as = "watch"}
        }
      })
    end)

    test("from LiteralStr use Name as Name, Name as Name", function()
      parser:parse([[from "fs" use FSWatcher as Watcher, watch as wtch]])
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
    end)

    test("from LiteralStr use (\\n  Name as Name\\n  Name as Name\n)", function(params)
      parser:parse(params[1])
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
      assert(parser:next()):isNil()
    end):iter(
      {subtitle=" # newline as separator", params='from "fs" use (\n  FSWatcher as Watcher\n  watch as wtch\n)'},
      {subtitle=" # comma as separator", params='from "fs" use (\n  FSWatcher as Watcher,\n  watch as wtch\n)'}
    )

    test("from LiteralStr use (\\n  Name as Name\\n  Name as Name)", function()
      parser:parse('from "fs" use (\n  FSWatcher as Watcher\n  watch as wtch)')
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "fs",
        members = {
          {name = "FSWatcher", as = "Watcher"},
          {name = "watch", as = "wtch"}
        }
      })
      assert(parser:next()):isNil()
    end)

    test("from LiteralStr use Name => {variables}", function()
      parser:parse('from "module" use mem1 as MEM1 => {var1,var2}, mem2 => {var3,var4}\nabc')
      assert(parser:next()):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.FROM,
        line = 1,
        col = 1,
        module = "module",
        members = {
          {name="mem1", as="MEM1", uVars={"var1","var2"}},
          {name="mem2", as="mem2", uVars={"var3","var4"}}
        }
      })
      assert(parser:next()):isTable():has({
        type = SentType.EXP,
        line = 2,
        col = 1
      })
    end)

    suite("errors", function()
      test("name expected after comma", function()
        parser:parse('from "fs" use\n  FSWatcher as Watcher,\n  watch as wtch')
        assert(function() parser:next() end):raises("anonymous code (1, 14): name expected.")
      end)
    end)
  end):tags("from")

  ----------------
  -- nextEnum() --
  ----------------
  suite("nextEnum()", function()
    test("enum - error", function()
      parser:parse("enum")
      assert(function() parser:next() end):raises("anonymous code (1, 5): name expected.")
    end)

    test("visibility enum Color", function(params)
      parser:parse(params[1] .. " enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = params[1],
        name = "Color",
        items = {}
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("enum Color", function()
      parser:parse("enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = nil,
        name = "Color",
        items = {}
      })
    end)

    test("enum Color RED - error", function()
      parser:parse("enum Color RED")
      assert(function() parser:next() end):raises("anonymous code (1, 12): invalid token.")
    end)

    test("visibility enum Color", function(params)
      parser:parse(params[1] .. " enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = params[1],
        name = "Color",
        items = {}
      })
    end):iter(
      {subtitle = "# export", params = "export"},
      {subtitle = "# pub", params = "pub"}
    )

    test("pub enum Color", function()
      parser:parse("pub enum Color")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        visib = "pub",
        name = "Color",
        items = {}
      })
    end)

    test("enum Color {RED}", function()
      parser:parse("enum Color {RED}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color {RED, GREEN}", function()
      parser:parse("enum Color {RED, GREEN}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}, {name = "GREEN", value = 2}}
      })
    end)

    test("enum Color {RED, GREEN, BLUE}", function()
      parser:parse("enum Color {RED, GREEN, BLUE}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 1},
          {name = "GREEN", value = 2},
          {name = "BLUE", value = 3}
        }
      })
    end)

    test("enum Color {RED = 12}", function()
      parser:parse("enum Color {RED = 12}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}}
      })
    end)

    test("enum Color {RED = 12, GREEN = 34}", function()
      parser:parse("enum Color {RED = 12, GREEN = 34}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}, {name = "GREEN", value = 34}}
      })
    end)

    test("enum Color {RED = 12, GREEN = 34, BLUE = 56}", function()
      parser:parse("enum Color {RED = 12, GREEN = 34, BLUE = 56}")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)

    test("enum Color\\n  RED", function()
      parser:parse("enum Color\n  RED")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color\\nRED", function()
      parser:parse("enum Color\nRED")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {}
      })
    end)

    test("enum Color\\n  RED\\n  GREEN", function()
      parser:parse("enum Color\n  RED\n  GREEN")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}, {name = "GREEN", value = 2}}
      })
    end)

    test("enum Color\\n  RED\\nGREEN", function()
      parser:parse("enum Color\n  RED\nGREEN")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 1}}
      })
    end)

    test("enum Color\\n  RED\\n  GREEN\\n  BLUE", function()
      parser:parse("enum Color\n  RED\n  GREEN\n  BLUE")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 1},
          {name = "GREEN", value = 2},
          {name = "BLUE", value = 3}
        }
      })
    end)

    test("enum Color\\n  RED = 12", function()
      parser:parse("enum Color\n  RED = 12")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}}
      })
    end)

    test("enum Color\\n  RED = 12\\n  GREEN = 34", function()
      parser:parse("enum Color\n  RED = 12\n  GREEN = 34")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {{name = "RED", value = 12}, {name = "GREEN", value = 34}}
      })
    end)

    test("enum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("enum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)

    test("@annot1 @annot2\\nenum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("@annot1 @annot2\nenum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
      assert(stmt.annots:has("annot1")):eq(true)
      assert(stmt.annots:has("annot2")):eq(true)
    end)

    test("@annot1\\n@annot2\\nenum Color\\n  RED = 12\\n  GREEN = 34\\n  BLUE = 56", function()
      parser:parse("@annot1\n@annot2\nenum Color\n  RED = 12\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
      assert(stmt.annots:has("annot1")):eq(true)
      assert(stmt.annots:has("annot2")):eq(true)
    end)

    test("enum with comments whitin elements", function()
      parser:parse("enum Color\n  #comment\nRED = 12\n\n#comment\n  GREEN = 34\n  BLUE = 56")
      stmt = parser:next()
      assert(stmt):isTable():has({
        type = SentType.STMT,
        subtype = StmtType.ENUM,
        name = "Color",
        items = {
          {name = "RED", value = 12},
          {name = "GREEN", value = 34},
          {name = "BLUE", value = 56}
        }
      })
    end)
  end):tags("enum")

  -----------
  -- while --
  -----------
  suite("while", function()
    test("while Exp do Exp", function()
      parser:parse("while true do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp; Exp do Exp", function()
      parser:parse("while x < 100; x += 1 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.iter:__tostring()):eq("(+= x 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do", function()
      parser:parse("while true do")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):isEmpty()
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence", function()
      parser:parse("while true do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp; Exp do\\n Sentence", function()
      parser:parse("while x < 100; x += 1 do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.iter:__tostring()):eq("(+= x 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\n Sentence", function()
      parser:parse("while true do\n print(x)\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.body[2]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\nSentence", function()
      parser:parse("while true do\n print(x)\nprint(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print y)")
      assert(parser:next()):isNil()
    end)

    test("while Exp do Block else Block", function()
      parser:parse("while true do\n print(x)\nelse\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n Sentence\\ncatch\\n  Sentence", function()
      parser:parse("while true do\n print(x)\ncatch\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch\\nprint(e)", function()
      parser:parse("while true do\n print(x)\ncatch\nprint(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(0)
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print e)")
      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch e\\n  print(e)", function()
      parser:parse("while true do\n  print(x)\ncatch e\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\nfinally\\n  print(y)", function()
      parser:parse("while true do\n print(x)\nfinally\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print y)")

      assert(parser:next()):isNil()
    end)

    test("while Exp do\\n print(x)\\ncatch e\\n print(y)\\nfinally\\n print(z)", function()
      parser:parse("while true do\n print(x)\ncatch e\n print(y)\nfinally\n print(z)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.WHILE
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")

      assert(parser:next()):isNil()
    end)
  end):tags("while")

  -----------
  -- until --
  -----------
  suite("until", function()
    test("until Exp do Exp", function()
      parser:parse("until true do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp; Exp do Exp", function()
      parser:parse("until x < 100; x += 1 do print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.iter:__tostring()):eq("(+= x 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do", function()
      parser:parse("until true do")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):isEmpty()
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence", function()
      parser:parse("until true do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp; Exp do\\n Sentence", function()
      parser:parse("until x < 100; x += 1 do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("(< x 100)")
      assert(stmt.iter:__tostring()):eq("(+= x 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\n Sentence", function()
      parser:parse("until true do\n print(x)\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.body[2]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\nSentence", function()
      parser:parse("until true do\n print(x)\nprint(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print y)")
      assert(parser:next()):isNil()
    end)

    test("until Exp do Block else Block", function()
      parser:parse("until true do\n print(x)\nelse\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print y)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n Sentence\\ncatch\\n  Sentence", function()
      parser:parse("until true do\n print(x)\ncatch\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch\\nprint(e)", function()
      parser:parse("until true do\n print(x)\ncatch\nprint(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):isNil()
      assert(stmt.catch.body):len(0)
      assert(stmt.finally):isNil()

      assert(parser:next():__tostring()):eq("(call print e)")
      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch e\\n  print(e)", function()
      parser:parse("until true do\n  print(x)\ncatch e\n  print(e)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print e)")
      assert(stmt.finally):isNil()

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\nfinally\\n  print(y)", function()
      parser:parse("until true do\n print(x)\nfinally\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isNil()
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print y)")

      assert(parser:next()):isNil()
    end)

    test("until Exp do\\n print(x)\\ncatch e\\n print(y)\\nfinally\\n print(z)", function()
      parser:parse("until true do\n print(x)\ncatch e\n print(y)\nfinally\n print(z)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.UNTIL
      })
      assert(stmt.cond:__tostring()):eq("true")
      assert(stmt.iter):isNil()
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable()
      assert(stmt.catch.var):eq("e")
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print y)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print z)")

      assert(parser:next()):isNil()
    end)
  end):tags("until")

  --------
  -- do --
  --------
  suite("do", function()
    test("do\\n Sentence", function()
      parser:parse("do\n print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
    end)

    test("do var name\\n Sentence", function()
      parser:parse("do var x\n print(y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        vars = {"x"},
        cond = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print y)")
    end)

    test("do var name, name\\n Sentence", function()
      parser:parse("do var x, y\n print(z)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        vars = {"x", "y"},
        cond = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print z)")
    end)

    test("do\\n Sentence\\ncatch\\n Sentence", function()
      parser:parse("do\n print(x)\ncatch\n x += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= x 1)")
    end)

    test("do\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\nfinally\n x += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
        catch = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= x 1)")
    end)

    test("do\\n Sentence\\ncatch\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\ncatch\n a += 1\nfinally\n b += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        cond = nil,
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= a 1)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= b 1)")
    end)

    test("do\\n Sentence\\nwhile Exp", function()
      parser:parse("do\n print(x)\nwhile x == 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.cond:__tostring()):eq("(== x 1)")
    end)

    test("do\\n Sentence\\nwhile Exp\\ncatch\\n Sentence\\nfinally\\n Sentence", function()
      parser:parse("do\n print(x)\nwhile x == 123\ncatch\n a += 1\nfinally\n b += 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.DO,
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print x)")
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(+= a 1)")
      assert(stmt.finally):isTable()
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(+= b 1)")
    end)
  end)

  ---------
  -- for --
  ---------
  suite("for", function()
    test("for Name ; Exp do Exp", function()
      parser:parse("for i; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        def = {
          {name = "i", value = nil}
        },
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name, Name ; Exp do Exp", function()
      parser:parse("for i, j; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        def = {
          {name = "i", value = nil},
          {name = "j", value = nil}
        },
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Name = Exp, Name = Exp ; Exp do Exp", function()
      parser:parse("for i = 0, j = 1; i < 10 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        iter = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(2)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(stmt.def[2].name):eq("j")
      assert(tostring(stmt.def[2].value)):eq("1")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Variable; Exp; Exp do Exp", function()
      parser:parse("for i = 0; i < 10; i += 1 do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(tostring(stmt.iter)):eq("(+= i 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for Variable; Exp; Exp do Block else Block", function()
      parser:parse('for i = 0; i < 10; i += 1 do print(i)\nelse print(j)')

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR,
        catch = nil,
        finally = nil
      })
      assert(stmt.def):len(1)
      assert(stmt.def[1].name):eq("i")
      assert(tostring(stmt.def[1].value)):eq("0")
      assert(tostring(stmt.cond)):eq("(< i 10)")
      assert(tostring(stmt.iter)):eq("(+= i 1)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print j)")
    end)
  end):tags("for")

  --------------
  -- for each --
  --------------
  suite("for each", function()
    test("error - :=, ::= or in expected", function()
      parser:parse("for each i list do print(i)")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ':=', '::=' or 'in' expected.")
    end)

    test("for each Name in Exp do Sentence", function(params)
      parser:parse("for each i " .. params[1] .. " list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # :=", params=":="}
    )

    test("for each Name ::= Exp do Sentence", function()
      parser:parse("for each i ::= list do print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "const",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for each Name, Name in Exp do Sentence", function(params)
      parser:parse("for each i, j " .. params[1] .. " list do print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end):iter(
      {subtitle=" # in", params="in"},
      {subtitle=" # :=", params=":="}
    )

    test("for each Name, Name ::= Exp do Sentence", function(params)
      parser:parse("for each i, j ::= list do print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "const",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end)

    test("for each Name in Exp do\\n Sentence", function()
      parser:parse("for each i in list do\n print(i)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = nil,
        value = "i",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
    end)

    test("for each Name, Name in Exp do\\n Sentence", function()
      parser:parse("for each i, j in list do\n print(i, j)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = "i",
        value = "j",
        el = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
    end)

    test("for each Name, Name in Exp do Block else Block", function()
      parser:parse("for each i, j in list do\n print(i, j)\nelse print(x)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = "i",
        value = "j",
        catch = nil,
        finally = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i j)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(call print x)")
    end)

    test("for each Name in Exp do\\n Sentence\\ncatch\\n Sentece\\nfinally\\n Sentence", function()
      parser:parse("for each i in list do\n print(i)\ncatch\n print(c)\nfinally\n print(f)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = SentType.STMT,
        subtype = StmtType.FOR_EACH,
        decl = "var",
        key = nil,
        value = "i",
        el = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(call print i)")
      assert(stmt.catch):isTable():has({
        var = nil
      })
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring()):eq("(call print c)")
      assert(stmt.finally.body):len(1)
      assert(stmt.finally.body[1]:__tostring()):eq("(call print f)")
    end)
  end)

  ----------
  -- proc --
  ----------
  suite("proc", function()
    test("proc Name() = Exp - error, = is invalid", function()
      parser:parse("proc sum(x, y) = x + y")
      assert(function() parser:next() end):raises()
    end)

    test("pr Name.Name() Body", function()
      parser:parse("pr Calcul.sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name:Name() Body", function()
      parser:parse("proc Calcul:sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "intl",
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name!Name() Body", function()
      parser:parse("proc Calcul!sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pvt",
        async = false,
        itype = "Calcul",
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("proc Name() Body", function()
      parser:parse("proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("export proc Name() Body", function()
      parser:parse("export proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("async proc Name() Body", function()
      parser:parse("async proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("pub async proc Name() Body", function()
      parser:parse("pub async proc sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)
  end)

  --------
  -- fn --
  --------
  suite("fn", function()
    suite("parameters", function()
      suite("=>", function()
        test("fn Name(Name=>[Name])", function()
          parser:parse("fn f(params=>[x])")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = "list",
                uVars = {"x"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>[Name,Name])", function()
          parser:parse("fn f(params=>[x,y])")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = "list",
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>{Name})", function()
          parser:parse("fn f(params=>{x})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = "map",
                uVars = {"x"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>{Name,Name})", function()
          parser:parse("fn f(params=>{x,y})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = "map",
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name:{...}=>{...})", function()
          parser:parse("fn f(params:{x:num}=>{x,y})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  {
                    type = "num",
                    name = "x",
                    mandatory = true
                  }
                },
                uVars = {"x", "y"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name=>) - error", function()
          parser:parse("fn f(params=>x)")
          assert(function() parser:next() end):raises("anonymous code (1, 14):")
        end)
      end)

      suite(":>", function()
        test("fn Name(Name:>[Name]) - { expected", function()
          parser:parse("fn f(params:>[x])")
          assert(function() parser:next() end):raises("anonymous code (1, 14): '{' expected.")
        end)

        test("fn Name(Name:>{Name})", function()
          parser:parse("fn f(params:>{x})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  {
                    type = "any",
                    name = "x",
                    mandatory = true
                  }
                },
                uVars = {"x"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)

        test("fn Name(Name:>{Name,Name})", function()
          parser:parse("fn f(params:>{a,b:num,c?:text})")

          stmt = parser:next()
          assert(stmt):isTable():has({
            line = 1,
            col = 1,
            visib = nil,
            itype = nil,
            name = "f",
            accessor = nil,
            rtype = nil,
            rvar = nil,
            params = {
              {
                const = false,
                optional = false,
                name = "params",
                type = {
                  {
                    type = "any",
                    name = "a",
                    mandatory = true
                  },
                  {
                    type = "num",
                    name = "b",
                    mandatory = true
                  },
                  {
                    type = "text",
                    name = "c",
                    mandatory = false
                  }
                },
                uVars = {"a", "b", "c"}
              }
            }
          })
          assert(stmt.body):isEmpty()
        end)
      end)

      test("fn Name ( Name )", function()
        parser:parse("fn inc(x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = nil,
              value = nil,
              uVars = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( const Name )", function()
        parser:parse("fn inc(const x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = true,
              modifier = nil,
              name = "x",
              optional = false,
              type = nil,
              value = nil,
              uVars = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name ? )", function()
        parser:parse("fn inc(x?)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name : Name )", function()
        parser:parse("fn inc(x:num)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = false,
              type = "num",
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name ?: Name )", function()
        parser:parse("fn inc(x?:num)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = "num",
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name ? : Name )", function()
        parser:parse("fn inc(x? :num)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          rtype = nil,
          rvar = nil,
          accessor = nil,
          params = {
            {
              const = false,
              modifier = nil,
              name = "x",
              optional = true,
              type = "num",
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( pub Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = ".",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # pub", params = "pub"},
        {subtitle = " # .", params = "."}
      )

      test("fn Name ( intl Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = ":",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # intl", params = "intl"},
        {subtitle = " # :", params = ":"}
      )

      test("fn Name ( pvt Name )", function(params)
        parser:parse(string.format("fn inc(%s x)", params[1]))

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          params = {
            {
              const = false,
              modifier = "!",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end):iter(
        {subtitle = " # pvt", params = "pvt"},
        {subtitle = " # !", params = "!"}
      )

      test("fn Name ( ... Name )", function()
        parser:parse("fn inc(...x)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          rtype = nil,
          rvar = nil,
          accessor = nil,
          params = {
            {
              const = false,
              modifier = "...",
              name = "x",
              optional = false,
              type = nil,
              value = nil
            }
          }
        })
        assert(stmt.body):isEmpty()
      end)

      test("fn Name(Name : {})", function()
        parser:parse("fn myfn(p:{})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {}
        })
      end)

      test("fn Name(Name : {Name})", function()
        parser:parse("fn myfn(p:{x})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = "any", mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name?})", function()
        parser:parse("fn myfn(p:{x?})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = "any", mandatory = false}
          }
        })
      end)

      test("fn Name(Name : {Name : Name})", function()
        parser:parse("fn myfn(p:{x:num})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = "num", mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name : (Name)})", function()
        parser:parse("fn myfn(p:{x:(text)})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = {"text"}, mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name : (Name, Name)})", function()
        parser:parse("fn myfn(p:{x:(text,list)})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = {"text", "list"}, mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name : [Name]})", function()
        parser:parse("fn myfn(p:{x:[text]})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = {"list*", "text"}, mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name : [Name, Name]})", function()
        parser:parse("fn myfn(p:{x:[text,num]})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = {"list*", "text", "num"}, mandatory = true}
          }
        })
      end)

      test("fn Name(Name : {Name ? : Name})", function()
        parser:parse("fn myfn(p:{x?:num})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = "num", mandatory = false}
          }
        })
      end)

      test("fn Name(Name : {Name, Name?:Name})", function()
        parser:parse("fn myfn(p:{x, y?:text})")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {}
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "p",
          optional = false,
          type = {
            {name = "x", type = "any", mandatory = true},
            {name = "y", type = "text", mandatory = false}
          }
        })
      end)

      test("fn Name (Name : (Name))", function()
        parser:parse("fn myfn(p:(text))")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = false,
              type = {"text"}
            }
          }
        })
      end)

      test("fn Name (Name : (Name, Name))", function()
        parser:parse("fn myfn(p:(text,list))")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = false,
              type = {"text", "list"}
            }
          }
        })
      end)

      test("fn Name (Name ?: (Name, Name))", function()
        parser:parse("fn myfn(p?:(text,list))")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = true,
              type = {"text", "list"}
            }
          }
        })
      end)

      test("fn Name (Name : [])", function()
        parser:parse("fn myfn(p:[])")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = true,
              type = "list*"
            }
          }
        })
      end)

      test("fn Name (Name : [Name])", function()
        parser:parse("fn myfn(p:[text])")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = false,
              type = {"list*", "text"}
            }
          }
        })
      end)

      test("fn Name (Name : [Name, Name])", function()
        parser:parse("fn myfn(p:[text,num])")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = false,
              type = {"list*", "text", "num"}
            }
          }
        })
      end)

      test("fn Name (Name ?: [Name, Name])", function()
        parser:parse("fn myfn(p?:[text,num])")
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "myfn",
          accessor = nil,
          rtype = nil,
          rvar = nil,
          body = {},
          params = {
            {
              const = false,
              modifier = nil,
              name = "p",
              optional = true,
              type = {"list*", "text", "num"}
            }
          }
        })
      end)

      test("fn Name ( Name := LiteralNum )", function()
        parser:parse("fn inc(x := 123)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "num",
        })
        assert(stmt.params[1].value:__tostring()):eq("123")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := LiteralText )", function()
        parser:parse('fn inc(x := "texto")')

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "text",
        })
        assert(stmt.params[1].value:__tostring()):eq("texto")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := true )", function()
        parser:parse("fn inc(x := true)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "bool",
        })
        assert(stmt.params[1].value:__tostring()):eq("true")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := false )", function()
        parser:parse("fn inc(x := false)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "bool",
        })
        assert(stmt.params[1].value:__tostring()):eq("false")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := list )", function()
        parser:parse("fn inc(x := [])")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "list",
        })
        assert(tostring(stmt.params[1].value)):eq("[]")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := map )", function()
        parser:parse("fn inc(x := {})")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          itype = nil,
          name = "inc",
          accessor = nil,
          rtype = nil,
          rvar = nil
        })
        assert(stmt.params):len(1)
        assert(stmt.params[1]):has({
          const = false,
          modifier = nil,
          name = "x",
          optional = true,
          type = "map",
        })
        assert(tostring(stmt.params[1].value)):eq("{}")
        assert(stmt.body):isEmpty()
      end)

      test("fn Name ( Name := nil ) - error", function()
        parser:parse("fn inc(x := nil)")
        assert(function() parser:next() end):raises("anonymous code (1, 13): for infering type, the default value must be a literal: bool, list, map, num or text.")
      end)

      test("fn Name . Name () : 123", function()
        parser:parse("fn MyType.method() : 123")
        assert(function() parser:next() end):raises("anonymous code (1, 22): name expected.")
      end)
    end)

    test("export async fn Name() Body", function()
      parser:parse("export async fn sum(x, y) x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(+ x y)")
    end)

    test("fn Name ( Name ? = Exp ) = Exp - error", function()
      parser:parse("fn inc(x ? = 123) = x + 1")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ')' expected.")
    end)

    test("fn Name ( Name = Exp ) = Exp", function()
      parser:parse("fn inc(x = 123) = x + 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = nil,
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x 1)")
    end)

    test("fn Name ( Name = Exp ) = Xml", function()
      parser:parse("#!react\nfn inc(x = 123) = <X/>")

      parser:next() --#!react

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = nil,
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (<X></X>)")
    end)

    test("fn Name ( Name : Name = Exp ) = Exp", function()
      parser:parse("fn inc(x : num = 123) = x + 1")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        itype = nil,
        name = "inc",
        accessor = nil,
        rtype = nil,
        rvar = nil
      })
      assert(stmt.params):len(1)
      assert(stmt.params[1]):has({
        const = false,
        modifier = nil,
        name = "x",
        optional = true,
        type = "num",
      })
      assert(stmt.params[1].value:__tostring()):eq("123")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x 1)")
    end)

    test("fn Name ( Name ? := Exp ) = Exp - error", function()
      parser:parse("fn inc(x ? := 123) = x + 1")
      assert(function() parser:next() end):raises("anonymous code (1, 12): ')' expected.")
    end)

    test("visibility fn Name ()", function(params)
      parser:parse(params[1] .. " fn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = params[1],
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("fn Name ()", function()
      parser:parse("fn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("@annot1\\nfn Name ()", function()
      parser:parse("@annot1\nfn myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
      assert(stmt.annots:has("annot1")):eq(true)
    end)

    test("fn Name () -> self", function()
      parser:parse("fn myfn() -> self")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = "self",
        params = {},
        body = {}
      })
    end)

    test("fn Name () -> Name", function()
      parser:parse("fn myfn() -> num")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = "num",
        params = {},
        body = {}
      })
    end)

    test("fn Name () -> keyword", function()
      parser:parse("fn myfn() -> type")
      assert(function() parser:next() end):raises("anonymous code (1, 14): return value must be 'self' or a name.")
    end)

    test("fn Name () -> Name : Name", function()
      parser:parse("fn myfn() -> res:num")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = "num",
        rvar = "res",
        params = {},
        body = {}
      })
    end)

    test("fn Name . Name ()", function()
      parser:parse("fn MyType.myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name : Name ()", function()
      parser:parse("fn MyType:myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "intl",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name ! Name ()", function()
      parser:parse("fn MyType!myfn()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pvt",
        async = false,
        itype = "MyType",
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {},
        body = {}
      })
    end)

    test("fn Name () = Exp", function()
      parser:parse("fn myfn() = 12 * 34")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (* 12 34)")
    end)

    test("fn Name () : Name = Exp", function()
      parser:parse("fn myfn() : num = 12 * 34")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "myfn",
        rtype = "num",
        rvar = nil,
        accessor = nil,
        params = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (* 12 34)")
    end)

    test("fn Name ( Name , Name ) = Exp", function()
      parser:parse("fn sum(x, y) = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = false,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("async fn Name() Body", function()
      parser:parse("async fn sum(x, y) = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ x y)")
    end)

    test("async fn Name() -> var Body", function()
      parser:parse("async fn sum(x, y) -> r\n  r = x + y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        async = true,
        itype = nil,
        name = "sum",
        accessor = nil,
        rtype = nil,
        rvar = nil,
        params = {
          {
            const = false,
            modifier = nil,
            name = "x",
            optional = false,
            type = nil,
            value = nil
          },
          {
            const = false,
            modifier = nil,
            name = "y",
            optional = false,
            type = nil,
            value = nil
          }
        }
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= r (+ x y))")
    end)
  end):tags("fn")

  ----------
  -- intf --
  ----------
  suite("intf", function()
    suite("errors", function()
      test(":type expected", function()
        parser:parse("export intf Param\n  x?")
        assert(function() parser:next() end):raises("anonymous code (2, 4): :type expected.")
      end)
    end)

    test("intf", function(params)
      parser:parse(params[1] .. " intf Param\n  x\n  y?:num\n  z:text")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = params[1],
        name = "Param",
        fields = {
          {name="x", optional=false, type=nil},
          {name="y", optional=true, type="num"},
          {name="z", optional=false, type="text"}
        }
      })
      assert(stmt.annots):len(0)

      assert(parser:next()):isNil()
    end):iter(
      {subtitle="export", params="export"},
      {subtitle="pub", params="pub"}
    )

    test("intf", function()
      parser:parse("intf Param\n  x")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        type = "stmt",
        subtype = "intf",
        visib = nil,
        name = "Param",
        fields = {
          {name="x", optional=false, type=nil}
        }
      })
      assert(stmt.annots):len(0)

      assert(parser:next()):isNil()
    end)
  end):tags("intf")

  ------------
  -- struct --
  ------------
  suite("struct", function()
    suite("errors", function()
      test("var or const expected", function()
        parser:parse("struct MyStruct\n  pub x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 7): 'const', 'var', 'async', 'fn' or 'proc' expected.")
      end)

      test("visibility expected", function()
        parser:parse("struct MyStruct\n  const x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 3): 'pub', 'intl' or 'pvt' expected.")
      end)

      test("field or method expcted", function()
        parser:parse("struct MyStruct\n  x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 3): field or method expected.")
      end)

      test("proc or fn expected", function()
        parser:parse("struct MyStruct\n  pub async x:num")
        assert(function() parser:next() end):raises("anonymous code (2, 13): 'proc' or 'fn' expected.")
      end)
    end)

    test("template", function()
      parser:parse("use Tmpl<T>\nexport struct MyStruct : Tmpl<MyT>\nx = 1", "./test/data/template/MyStruct.dog")
      assert(parser:next()):isTable():has({line=1, col=1, subtype=StmtType.USE})

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "MyStruct",
        base = nil
      })
      assert(stmt.annots._items["Self"]):isNotNil()
      assert(parser:next()):isTable():has({line=3, col=1, type=SentType.EXP})
    end)

    test("@emitter struct Name", function()
      parser:parse("@emitter struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 10,
        visib = nil,
        name = "MyStruct",
        base = "dogma.EventEmitter",
        fields = {}
      })
      assert(stmt.annots:has("emitter")):eq(true)
    end)

    test("export struct Name", function()
      parser:parse("export struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "export",
        name = "MyStruct",
        base = nil,
        fields = {}
      })
    end)

    test("pub struct Name", function()
      parser:parse("pub struct MyStruct")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = "pub",
        name = "MyStruct",
        base = nil,
        fields = {}
      })
    end)

    test("struct Name", function()
      parser:parse("struct Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        fields = {},
      })
    end)

    test("struct Name : Name", function()
      parser:parse("struct Coord3D : Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        fields = {},
      })
    end)

    suite("fields", function()
      test("const fields", function()
        parser:parse('struct MyStruct\n  pub const a\n  pvt const b:num\n  intl const c:=0\n  pub const d:num=0')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(4)
        assert(stmt.fields[1]):has({kind="const", visib="pub", name="a", optional=false, type=nil, value=nil})
        assert(stmt.fields[2]):has({kind="const", visib="pvt", name="b", optional=false, type="num", value=nil})
        assert(stmt.fields[3]):has({kind="const", visib="intl", name="c", optional=true, type="num"})
        assert(tostring(stmt.fields[3].value)):eq("0")
        assert(stmt.fields[4]):has({kind="const", visib="pub", name="d", optional=true, type="num"})
        assert(tostring(stmt.fields[4].value)):eq("0")
      end)

      test("var fields", function()
        parser:parse('struct MyStruct\n  pub var a\n  pvt var b?\n  intl var c:num\n  pub var d?:num\n  pub var e:num=0\n  pub var f:=0')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(6)
        assert(stmt.fields[1]):has({kind="var", visib="pub", name="a", optional=false, type=nil, value=nil})
        assert(stmt.fields[2]):has({kind="var", visib="pvt", name="b", optional=true, type=nil, value=nil})
        assert(stmt.fields[3]):has({kind="var", visib="intl", name="c", optional=false, type="num", value=nil})
        assert(stmt.fields[4]):has({kind="var", visib="pub", name="d", optional=true, type="num", value=nil})
        assert(stmt.fields[5]):has({kind="var", visib="pub", name="e", optional=true, type="num"})
        assert(tostring(stmt.fields[5].value)):eq("0")
        assert(stmt.fields[6]):has({kind="var", visib="pub", name="f", optional=true, type="num"})
        assert(tostring(stmt.fields[6].value)):eq("0")
      end)
    end)

    suite("methods", function()
      test("fn", function()
        parser:parse('struct MyStruct\n  pub async fn method1()\n  @abstract\n  pvt fn method2()\n  intl fn method3()')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(0)
        assert(stmt.methods):len(3)
        assert(stmt.methods[1]):has({
          name = "method1",
          itype = "MyStruct",
          visib = "pub",
          subtype = "fn",
          line = 2,
          col = 3,
          async = true
        })
        assert(stmt.methods[2]):has({
          name = "method2",
          itype = "MyStruct",
          visib = "pvt",
          subtype = "fn",
          line = 4,
          col = 3,
          async = false
        })
        assert(stmt.methods[2].annots._items["abstract"]):has({value="abstract"})
        assert(stmt.methods[3]):has({
          name = "method3",
          itype = "MyStruct",
          visib = "intl",
          subtype = "fn",
          line = 5,
          col = 3,
          async = false
        })
        assert(stmt.methods[3].annots._items["abstract"]):isNil()
      end)

      test("proc", function()
        parser:parse('struct MyStruct\n  pub async proc method1()\n  @abstract\n  pvt proc method2()\n  intl proc method3()')
        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "MyStruct",
          base = nil
        })
        assert(stmt.fields):len(0)
        assert(stmt.methods):len(3)
        assert(stmt.methods[1]):has({
          name = "method1",
          itype = "MyStruct",
          visib = "pub",
          subtype = "proc",
          line = 2,
          col = 3,
          async = true
        })
        assert(stmt.methods[2]):has({
          name = "method2",
          itype = "MyStruct",
          visib = "pvt",
          subtype = "proc",
          line = 4,
          col = 3,
          async = false
        })
        assert(stmt.methods[2].annots._items["abstract"]):has({value="abstract"})
        assert(stmt.methods[3]):has({
          name = "method3",
          itype = "MyStruct",
          visib = "intl",
          subtype = "proc",
          line = 5,
          col = 3,
          async = false
        })
        assert(stmt.methods[3].annots._items["abstract"]):isNil()
      end)
    end)
  end)

  ----------
  -- type --
  ----------
  suite("type", function()
    test("@emitter\\ntype Name()", function()
      parser:parse("@emitter\ntype Abc()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        name = "Abc",
        params = {},
        base = "dogma.EventEmitter",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.annots:has("emitter")):eq(true)
    end)

    test("@annot1\\ntype Name()", function()
      parser:parse("@annot1\ntype Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 2,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.annots:has("annot1")):eq(true)
    end)

    test("type Name", function()
      parser:parse("type Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = nil,
        catch = nil,
        finally = nil
      })
    end)

    test("type Name with fields", function()
      parser:parse('type Coord2D\n  pub const x:=0\n  var y?:num=0\n  pvt const ctype:="2D"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        bargs = nil,
        body = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):isTable():has({
        const = true,
        modifier = ".",
        name = "x",
        type = "num",
        optional = false,
      })
      assert(tostring(stmt.params[1].value)):eq("0")
      assert(stmt.params[2]):isTable():has({
        const = false,
        modifier = ".",
        type = "num",
        optional = true,
        value = nil
      })
      assert(tostring(stmt.params[2].value)):eq("0")
      assert(stmt.params[3]):isTable():has({
        const = true,
        modifier = ":",
        name = "ctype",
        type = "text",
        optional = false,
      })
      assert(tostring(stmt.params[3].value)):eq("2D")
    end)

    test("type Name - error: const or var expected", function()
      parser:parse('type Coord2D\n  pub x:=0\n  var y?:num=0\n  pvt const ctype:="2D"')
      assert(function() parser:next() end):raises("anonymous code (2, 7): var or const expected.")
    end)

    test("type Name()", function()
      parser:parse("type Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("visibility type Name()", function(params)
      parser:parse(params[1] .. " type Coord2D()")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = params[1],
        name = "Coord2D",
        params = {},
        base = nil,
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end):iter(
      {subtitle = " # export", params = "export"},
      {subtitle = " # pub", params = "pub"}
    )

    test("type Name(params)", function()
      parser:parse("type Coord2D(x, y)\n  .x = x\n  .y = y")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord2D",
        base = nil,
        bargs = nil,
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(2)
      assert(stmt.params[1]):has({name = "x"})
      assert(stmt.params[2]):has({name = "y"})
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring()):eq("(= (. x) x)")
      assert(stmt.body[2]:__tostring()):eq("(= (. y) y)")
    end)

    test("type Name() : Name", function()
      parser:parse("type Coord3D() : Coord2D")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        params = {},
        base = "Coord2D",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("type Name() : Name.Name", function()
      parser:parse("type Coord3D() : Coord.D2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        params = {},
        base = "Coord.D2",
        bargs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
    end)

    test("type Name(params) : Name(args)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(x, y)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):has({name = "x", modifier = nil})
      assert(stmt.params[2]):has({name = "y", modifier = nil})
      assert(stmt.params[3]):has({name = "z", modifier = "."})
      assert(stmt.bargs):len(2)
      assert(stmt.bargs[1]:__tostring()):eq("x")
      assert(stmt.bargs[2]:__tostring()):eq("y")
    end)

    test("type Name(params) : Name(...)", function()
      parser:parse("type Coord3D(x, y, .z) : Coord2D(...)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(3)
      assert(stmt.params[1]):has({name = "x", modifier = nil})
      assert(stmt.params[2]):has({name = "y", modifier = nil})
      assert(stmt.params[3]):has({name = "z", modifier = "."})
      assert(stmt.bargs):len(1)
      assert(stmt.bargs[1]):eq("...")
    end)

    test("type Name(...) : Name(...)", function()
      parser:parse("type Coord3D(...) : Coord2D(...)")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        visib = nil,
        name = "Coord3D",
        base = "Coord2D",
        mixin = nil,
        margs = nil,
        body = {},
        catch = nil,
        finally = nil
      })
      assert(stmt.params):len(0)
      assert(stmt.bargs):len(1)
      assert(stmt.bargs[1]):eq("...")
    end)

    suite("mixin", function()
      test("@mixable type Name", function()
        parser:parse("@mixable type Startable")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 10,
          visib = nil,
          name = "Startable",
          base = nil,
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.annots:has("mixable")):eq(true)
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name :: Mixin()", function()
        parser:parse("type Server() :: Startable()")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          margs = {},
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name() :: Mixin(...)", function()
        parser:parse("type Server() :: Startable(...)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          margs = {"..."},
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
      end)

      test("type Name :: Mixin(args)", function()
        parser:parse("type Server() :: Startable(one, two)")

        stmt = parser:next()
        assert(stmt):isTable():has({
          line = 1,
          col = 1,
          visib = nil,
          name = "Server",
          base = nil,
          bargs = nil,
          mixin = "Startable",
          body = nil,
          catch = nil,
          finally = nil
        })
        assert(stmt.params):len(0)
        assert(stmt.bargs):isNil()
        assert(stmt.margs):len(2)
        assert(tostring(stmt.margs[1])):eq("one")
        assert(tostring(stmt.margs[2])):eq("two")
      end)
    end):tags("mixin")
  end):tags("type")

  -----------
  -- async --
  -----------
  suite("async", function()
    test("async Exp", function()
      parser:parse("async 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {},
        catch = nil
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {invalid=Exp} Exp", function()
      parser:parse("async with {invalid=250+250} 1+2")
      assert(function() parser:next() end):raises("delay or repeat expected.")
    end)

    test("async with {delay=Exp} Exp", function()
      parser:parse("async with {delay=250+250} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("(+ 250 250)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {delay=Exp, repeat=Exp} Exp", function()
      parser:parse("async with {delay=250+250, repeat=10+20} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("(+ 250 250)")
      assert(tostring(stmt.opts["repeat"])):eq("(+ 10 20)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async with {delay} Exp", function()
      parser:parse("async with {delay} 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        catch = nil
      })
      assert(tostring(stmt.opts.delay)):eq("delay")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
    end)

    test("async\\n Exp\\n Exp", function()
      parser:parse("async\n 1+2\n 3+4")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {},
        catch = nil
      })
      assert(stmt.body):len(2)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
      assert(stmt.body[2]:__tostring(("+ 3 4")))
    end)

    test("async\\n Exp\\ncatch\\n Exp", function()
      parser:parse("async\n 1+2\ncatch\n 3+4")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.ASYNC,
        opts = {}
      })
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring(("+ 1 2")))
      assert(stmt.catch.body):len(1)
      assert(stmt.catch.body[1]:__tostring(("+ 3 4")))
    end)
  end):tags("async")

  --------
  -- if --
  --------
  suite("if", function()
    test("if Exp then Exp", function()
      parser:parse("if x == 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if name := Exp then Exp", function()
      parser:parse("if x := 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if name ::= Exp then Exp", function()
      parser:parse("if x ::= 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "const",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if var Exp then Exp", function()
      parser:parse("if var x = 123 then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if var Exp; then Exp", function()
      parser:parse("if var x = 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "var",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if const Exp; then Exp", function()
      parser:parse("if const x = 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = "const",
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(= x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp; Exp then Exp", function()
      parser:parse("if x == 123; x then x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(tostring(stmt.value)):eq("(== x 123)")
      assert(tostring(stmt.cond)):eq("x")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp then Stmt", function()
      parser:parse("if x == 123 then return 1+2")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("return (+ 1 2)")
    end)

    test("if Exp then Exp else Exp", function()
      parser:parse("if x == 123 then x = 321 else x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then Exp\\nelse Exp", function()
      parser:parse("if x == 123 then x = 321\nelse x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then\\n Exp", function()
      parser:parse("if x == 123 then\n x = 321")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
    end)

    test("if Exp then\\n Exp\\nelse\\n Exp", function()
      parser:parse("if x == 123 then\n x = 321\nelse\n x = 135")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        elif = nil,
      })
      assert(stmt.cond:__tostring()):eq("(== x 123)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 321)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 135)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp\\nelse if Exp then\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444\nelse if x == 555 then\n x = 666")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(2)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.elif[2].cond:__tostring()):eq("(== x 555)")
      assert(stmt.elif[2].body):len(1)
      assert(stmt.elif[2].body[1]:__tostring()):eq("(= x 666)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp\\nelse if Exp then Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444\nelse if x == 555 then x = 666")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil,
        el = nil
      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(2)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.elif[2].cond:__tostring()):eq("(== x 555)")
      assert(stmt.elif[2].body):len(1)
      assert(stmt.elif[2].body[1]:__tostring()):eq("(= x 666)")
    end)

    test("if Exp then\\n Exp\\nelse if Exp then\\n Exp\\nelse\\n Exp", function()
      parser:parse("if x == 111 then\n x = 222\nelse if x == 333 then\n x = 444\nelse x = 555")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 555)")
    end)

    test("if Exp then Exp\\nelse if Exp then Exp\\nelse Exp", function()
      parser:parse("if x == 111 then x = 222\nelse if x == 333 then x = 444\nelse x = 555")

      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):len(1)
      assert(stmt.el[1]:__tostring()):eq("(= x 555)")
    end)

    test("  if Exp then\\n Exp\\nelse", function()
      parser:parse("  if x == 111 then\n    x = 222\nelse\n  x = 333")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 3,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):isNil()
      assert(stmt.el):isNil()
    end)

    test("  if Exp then\\n Exp\\n  else if Exp then\\n    Exp\\nelse", function()
      parser:parse("  if x == 111 then\n    x = 222\n  else if x == 333 then\n    x = 444\nelse\n  x = 555")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 3,
        subtype = StmtType.IF,
        decl = nil

      })
      assert(stmt.cond:__tostring()):eq("(== x 111)")
      assert(stmt.body):len(1)
      assert(stmt.body[1]:__tostring()):eq("(= x 222)")
      assert(stmt.elif):len(1)
      assert(stmt.elif[1].cond:__tostring()):eq("(== x 333)")
      assert(stmt.elif[1].body):len(1)
      assert(stmt.elif[1].body[1]:__tostring()):eq("(= x 444)")
      assert(stmt.el):isNil()
    end)
  end):tags("if")

  ---------
  -- pub --
  ---------
  suite("pub", function()
    suite("error", function()
      test("pub - error, expecting text or name", function()
        parser:parse("pub")
        assert(function() parser:next() end):raises("anonymous code (1, 4): literal text or name expected.")
      end)

      test("pub Name, - error, expecting text or name", function()
        parser:parse("pub x,")
        assert(function() parser:next() end):raises("anonymous code (1, 7): literal text or name expected.")
      end)

      test("pub Name Name - error", function()
        parser:parse("pub Item1 Item2")
        assert(function() parser:next() end):raises("anonymous code (1, 11): comma or newline expected.")
      end)

      test("pub 123 - error", function()
        parser:parse("pub 123")
        assert(function() parser:next() end):raises("anonymous code (1, 5): literal text or name expected.")
      end)
    end)

    test("pub Name", function()
      parser:parse("pub Item")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "pub", value = "Item"}}
      })
    end)

    test("pub Text", function()
      parser:parse('pub "Item"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "use", value = {path = "Item", name = "Item"}}}
      })
    end)

    test("pub Name, Text", function()
      parser:parse('pub Item1, "Item2"')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {{type = "pub", value = "Item1"}, {type = "use", value = {path = "Item2", name = "Item2"}}}
      })
    end)

    test("pub (\\nName1, Name2\\nText\\n)", function()
      parser:parse('pub (\nItem1, Item2\n"Item3"\n)')
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.PUB,
        items = {
          {type="pub", value="Item1"},
          {type="pub", value="Item2"},
          {type="use", value={path="Item3", name="Item3"}}
        }
      })
    end)
  end):tags("pub")

  ------------------
  -- nextExport() --
  ------------------
  suite("nextExport()", function()
    test("export", function()
      parser:parse("export")
      assert(function() parser:next() end):raises("anonymous code (1, 7): incomplete expression.")
    end)

    test("export Exp", function()
      parser:parse("export 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.EXPORT
      })
      assert(tostring(stmt.exp)):eq("(+ 1 2)")
    end)
  end):tags("export")

  ----------------
  -- nextWith() --
  ----------------
  suite("nextWith()", function()
    test("with", function()
      parser:parse("with")
      assert(function() parser:next() end):raises("anonymous code (1, 5): incomplete expression.")
    end)

    test("with Exp", function()
      parser:parse("with 1+2")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH,
        ifs = {},
        els = nil
      })
      assert(parser:next()):isNil()
    end)

    test("with Exp if Exp then - with one sentence", function()
      parser:parse("with 1+2\n  if 1 then x")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(1)
      assert(tostring(stmt.ifs[1].cond)):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(parser:next()):isNil()
    end)

    test("with Exp if Exp then - with several sentences", function()
      parser:parse("with 1+2\n  if 1 then x\n    y")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(1)
      assert(tostring(stmt.ifs[1].cond)):eq("1")
      assert(stmt.ifs[1].body):len(2)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[1].body[2])):eq("y")
      assert(parser:next()):isNil()
    end)

    test("with Exp - with several if", function()
      parser:parse("with 1+2\n  if 1 then x\n  if 2 then y")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH,
        els = nil
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].cond)):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[2].cond)):eq("2")
      assert(stmt.ifs[2].body):len(1)
      assert(tostring(stmt.ifs[2].body[1])):eq("y")
      assert(parser:next()):isNil()
    end)

    test("with Exp - with if and else", function()
      parser:parse("with 1+2\n  if 1 then x\n  if 2 then y\n  else z\n")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].cond)):eq("1")
      assert(stmt.ifs[1].body):len(1)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[2].cond)):eq("2")
      assert(stmt.ifs[2].body):len(1)
      assert(tostring(stmt.ifs[2].body[1])):eq("y")
      assert(stmt.els):isNotNil()
      assert(stmt.els):len(1)
      assert(tostring(stmt.els[1])):eq("z")
      assert(parser:next()):isNil()
    end)

    test("full with", function()
      parser:parse("with 1+2\n  if 1 then x\n    y\n  if 2 then a\n    b\n  else 1\n    2\nnop")
      stmt = parser:next()
      assert(stmt):isTable():has({
        line = 1,
        col = 1,
        subtype = StmtType.WITH
      })
      assert(tostring(stmt.value)):eq("(+ 1 2)")
      assert(stmt.ifs):len(2)
      assert(tostring(stmt.ifs[1].cond)):eq("1")
      assert(stmt.ifs[1].body):len(2)
      assert(tostring(stmt.ifs[1].body[1])):eq("x")
      assert(tostring(stmt.ifs[1].body[2])):eq("y")
      assert(tostring(stmt.ifs[2].cond)):eq("2")
      assert(stmt.ifs[2].body):len(2)
      assert(tostring(stmt.ifs[2].body[1])):eq("a")
      assert(tostring(stmt.ifs[2].body[2])):eq("b")
      assert(stmt.els):isNotNil()
      assert(stmt.els):len(2)
      assert(tostring(stmt.els[1])):eq("1")
      assert(tostring(stmt.els[2])):eq("2")
      assert(parser:next()):isTable():has({
        line = 8,
        col = 1,
        type = StmtType.EXP
      })
    end)
  end):tags("with")
end):tags("stmt")
